﻿/*
 * Author: G. Monz (DK7IO), 2019-03-08
 * This file is distributed without any warranty.
 * */

using System.Text;
using System.IO;

namespace GM.Geodesy
{
	/// <summary>
	/// Creates CSV files (Comma Separated Values).
	/// </summary>
	public static class CSV
	{
		/// <summary>
		/// Saves a CSV file for a given 'Maidenhead Locator' with its subgrids.
		/// </summary>
		/// <param name="maidenheadLocator">The 'Maidenhead Locator'.</param>
		/// <param name="fileName">
		/// The full name of the CSV file with extension.
		/// The directory must exist with write permissions.
		/// </param>
		/// <param name="utf8">True if the resulted file will be UTF-8 encoded. False: ASCII (7 Bit).</param>
		/// <param name="angleFormat">A <see cref="AngleFormat" /> member.</param>
		/// <param name="decimalPlaces">The number of decimal places (after decimal point) for the last unit.</param>
		public static void SaveCSV(string maidenheadLocator,
			string fileName, bool utf8,
			AngleFormat angleFormat, int decimalPlaces)
		{
			string cellSeparator = ";";

			maidenheadLocator = (maidenheadLocator == null ? string.Empty : maidenheadLocator = maidenheadLocator.Trim());

			var hLocators = MaidenheadLocator.GetSubGrids(maidenheadLocator, MaidenheadLocator.SubGridsFilter.Left, true, true);
			int hLocatorsCount = hLocators.Count;

			var wLocators = MaidenheadLocator.GetSubGrids(maidenheadLocator, MaidenheadLocator.SubGridsFilter.Bottom, true, false);
			int wLocatorsCount = wLocators.Count;

			int linesCount = hLocatorsCount * 2 + 2;
			var lines = new string[linesCount];

			{
				bool forceAsciiCharacters = !utf8;

				for (int h = 0; h < hLocatorsCount; h++)
				{
					string hLocator = hLocators[h];
					string hDigit = hLocator.Substring(hLocator.Length - 1, 1);

					for (int w = 0; w < wLocatorsCount; w++)
					{
						string wLocator = wLocators[w];
						string wDigit = wLocator.Substring(wLocator.Length - 2, 1);

						if (h == 0)
						{
							string workLocator = maidenheadLocator + wDigit + hDigit;
							FillLines(lines, forceAsciiCharacters, cellSeparator, h, workLocator, true, angleFormat, decimalPlaces, PositionInRectangle.TopLeft);
						}

						{
							string workLocator = maidenheadLocator + wDigit + hDigit;
							FillLines(lines, forceAsciiCharacters, cellSeparator, h + 1, workLocator, true, angleFormat, decimalPlaces, PositionInRectangle.BottomLeft);
						}

						if (w == wLocatorsCount - 1)
						{
							string workLocator = maidenheadLocator + wDigit + hDigit;
							FillLines(lines, forceAsciiCharacters, cellSeparator, h, workLocator, false, angleFormat, decimalPlaces, PositionInRectangle.TopRight);

							if (h == hLocatorsCount - 1)
								FillLines(lines, forceAsciiCharacters, cellSeparator, h + 1, workLocator, false, angleFormat, decimalPlaces, PositionInRectangle.BottomRight);
						}
					}
				}
			}

			TrimLines(lines, linesCount);

			WriteFile(lines, linesCount, fileName, utf8);
		}

		static void FillLines(string[] lines, bool forceAsciiCharacters, string separator, int h, string subLocator, bool printSubLocator, AngleFormat angleFormat, int decimalPlaces, PositionInRectangle positionInRectangle)
		{
			var geographicalPoint = new GeographicalPoint(subLocator, positionInRectangle);

			int index = h * 2;

			lines[index] +=
				geographicalPoint.ToString(angleFormat, decimalPlaces, " / ", forceAsciiCharacters) +
				separator + separator;

			if (printSubLocator)
				lines[index + 1] += separator + subLocator + separator;
		}

		static void TrimLines(string[] lines, int linesCount)
		{
			for (int index = 0; index < linesCount; index++)
			{
				bool found = true;

				while (found)
				{
					string line = lines[index];
					found = (line.Substring(line.Length - 1) == ";");

					if (found)
						lines[index] = line.Substring(0, line.Length - 1);
				}
			}
		}

		static void WriteFile(string[] lines, int linesCount, string fileName, bool utf8)
		{
			var writer = new StreamWriter(
				fileName,
				false,
				(utf8 ? Encoding.UTF8 : Encoding.ASCII)
			);
			{
				for (int index = 0; index < linesCount; index += 2)
				{
					if (index != 1)
					{
						writer.WriteLine(lines[index]);

						if (index + 3 < linesCount)
							writer.WriteLine(lines[index + 3]);
					}
				}
			}
			writer.Close();
		}
	}
}