/*
 * Author: G. Monz (DK7IO), 2022-06-20
 * This file is distributed without any warranty.
 * */

using GM.Geodesy;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using QthManager.Properties;

namespace QthManager
{
	public partial class MainForm : Form
	{
		const string ProductText = "QTH Manager (by DK7IO, Version 2022-06-20)";

		const int MaxCsvDecimalPlaces = 7;

		string NewLine => Environment.NewLine;

		static void SetBusyCursor() => Cursor.Current = Cursors.WaitCursor;

		static void SetDefaultCursor() => Cursor.Current = Cursors.Default;

		KML.FileFormat KmlFileFormat
		{
			get
			{
				foreach (Control control in GroupBoxKmlFile.Controls)
				{
					if (!(control is RadioButton radioButton))
						continue;

					if (!radioButton.Checked)
						continue;

					if (!(control.Tag is string tag))
						continue;

					var converter = TypeDescriptor.GetConverter(typeof(KML.FileFormat));
					return (KML.FileFormat)converter.ConvertFromInvariantString(tag);
				}

				throw new InvalidOperationException("No valid KML File format.");
			}
		}

		AngleFormat AngleFormat
		{
			get
			{
				var listControlItem = comboBoxCsvAngleFormat.SelectedItem as ListControlItem;
				return (AngleFormat)listControlItem.Data;
			}
		}

		public MainForm()
		{
			InitializeComponent();

			Text = ProductText;

			radioButtonLocatorPositionInGridMiddleMiddle1.Checked = radioButtonLocatorPositionInGridMiddleMiddle2.Checked = true;

			CheckBoxSmallLettersForSubsquares.Checked = Settings.Default.SmallLettersForSubsquares;

			CheckBoxOpenFile.Checked = Settings.Default.OpenFile;

			ResetState1();
			ResetState2();

			TextBoxOutputPath.Text = string.IsNullOrEmpty(Settings.Default.OutputPath) ?
				Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) :
				Settings.Default.OutputPath;

			textBoxWaypointsCount1.Text = textBoxWaypointsCount2.Text = Settings.Default.WaypointsCount.ToString();
			textBoxWaypointsKm1.Text = textBoxWaypointsKm2.Text = Settings.Default.WaypointsKm.ToString();
			textBoxWaypointsAzimuth1.Text = textBoxWaypointsAzimuth2.Text = Settings.Default.WaypointsAzimuth.ToString();

			TextBoxLocator1.Text = TextBoxLocator2.Text = null;
			LocatorTextBoxChanged(TextBoxLocator1.Text, ButtonLocatorEnter1,
				ButtonSaveKml1, ButtonSaveCsv1, ButtonWaypoints1);
			LocatorTextBoxChanged(TextBoxLocator2.Text, ButtonLocatorEnter2,
				ButtonSaveKml2, ButtonSaveCsv2, ButtonWaypoints2);

			InitializeCsv();
			InitializeKml();
		}

		void InitializeCsv()
		{
			CheckBoxCsvUtf8.Checked = Settings.Default.CsvUtf8;

			TextBoxCsvDecimalPlaces.Text = Settings.Default.CsvDecimalPlaces.ToString();
			labelCsvDecimalPlaces.Text = "Decimal places (0..." + MaxCsvDecimalPlaces + ")";
			RefreshComboBoxCsvAngleFormat(keepSelectedIndex: false);

			SaveCsvContextMenu1.ItemClicked += new ToolStripItemClickedEventHandler(SaveCsvContextMenu1_ItemClicked);
			SaveCsvContextMenu2.ItemClicked += new ToolStripItemClickedEventHandler(SaveCsvContextMenu2_ItemClicked);
		}

		void InitializeKml()
		{
			RadioButtonKmlOutputKMZ.Checked = true;

			TextBoxKmlMinLodPixels.Text = Settings.Default.KmlMinLodPixel.ToString();

			TextBoxKmlColorEvenPrecisions.Text = Settings.Default.KmlColorEvenPrecisions;
			TextBoxKmlColorOddPrecisions.Text = Settings.Default.KmlColorOddPrecisions;

			SaveKmlContextMenu1.ItemClicked += new ToolStripItemClickedEventHandler(SaveKmlContextMenu1_ItemClicked);
			SaveKmlContextMenu2.ItemClicked += new ToolStripItemClickedEventHandler(SaveKmlContextMenu2_ItemClicked);

			CheckBoxKmlSuperGrid.Checked = Settings.Default.KmlSuperGrid;
			CheckBoxKmlAllSuperGrids.Checked = Settings.Default.KmlAllSuperGrids;
			SupergridCheckBoxChanged();

			CheckBoxKmlSubGrids.Checked = Settings.Default.KmlSubGrids;
		}

		void ButtonSummary_Click(object sender, EventArgs e)
		{
			try
			{
				var geographicalPoint1 = new GeographicalPoint(TextBoxLocator1.Text.Trim(), LocatorPositionInGrid(groupBoxLocatorPositionInGrid1));
				var geographicalPoint2 = new GeographicalPoint(TextBoxLocator2.Text.Trim(), LocatorPositionInGrid(groupBoxLocatorPositionInGrid2));

				var form = new SummaryForm(this);
				form.ShowDialog();
			}
			catch (ArgumentException)
			{
			}
		}

		void RefreshComboBoxCsvAngleFormat(bool keepSelectedIndex)
		{
			var angleFormatItems = new List<ListControlItem>();
			{
				var geographicalPoint = new GeographicalPoint(
					-4 + Math.Sqrt(2),
					+7 + Math.Sqrt(3));

				bool decimalPlacesOK = int.TryParse(TextBoxCsvDecimalPlaces.Text, out int decimalPlaces);

				if (decimalPlacesOK)
				{
					foreach (AngleFormat angleFormat in Enum.GetValues(typeof(AngleFormat)))
					{
						string displayText = geographicalPoint.ToString(angleFormat, decimalPlaces, " / ", false);
						var listControlItem = new ListControlItem(angleFormat, displayText);
						angleFormatItems.Add(listControlItem);
					}
				}
			}
			int selectedIndex = keepSelectedIndex ? comboBoxCsvAngleFormat.SelectedIndex : angleFormatItems.Count - 1;
			comboBoxCsvAngleFormat.DataSource = angleFormatItems;
			comboBoxCsvAngleFormat.SelectedIndex = selectedIndex;
		}

		void ShowSaveCsvContextMenu(MouseEventArgs mouseEventArgs, Control control, ToolStripDropDown contextMenu, string locator)
		{
			locator = locator.Trim();

			var items = contextMenu.Items;
			{
				items.Clear();

				{
					var item = new ToolStripMenuItem("Save CSV for '" + locator + "'");
					{
						item.Tag = new List<string>(new string[] { locator });
						item.Image = Properties.Resources.go;
					}
					items.Add(item);
				}

				if (locator.Length / 2 < MaidenheadLocator.MaxPrecision)
				{
					var subLocators = MaidenheadLocator.GetSubGrids(locator, MaidenheadLocator.SubGridsFilter.All, CheckBoxSmallLettersForSubsquares.Checked, false);
					string firstSubLocator = subLocators[0];
					string lastSubLocator = subLocators[subLocators.Count - 1];

					var item = new ToolStripMenuItem("Save CSV for '" + firstSubLocator + "' ... '" + lastSubLocator + "' (and do not open files)");
					{
						item.Tag = subLocators;
						item.Image = Properties.Resources.fast_forward;
					}
					items.Add(item);
				}
			}

			contextMenu.Show(control, mouseEventArgs.X, mouseEventArgs.Y);
		}

		void ShowSaveKmlContextMenu(MouseEventArgs mouseEventArgs, Control control, ToolStripDropDown contextMenu, string locator)
		{
			locator = locator.Trim();

			var items = contextMenu.Items;
			{
				items.Clear();

				{
					var item = new ToolStripMenuItem("Save KML for '" + locator + "'");
					{
						item.Tag = new List<string>(new string[] { locator });
						item.Image = Properties.Resources.go;
					}
					items.Add(item);
				}

				if (locator.Length / 2 < MaidenheadLocator.MaxPrecision)
				{
					var subLocators = MaidenheadLocator.GetSubGrids(locator, MaidenheadLocator.SubGridsFilter.All, CheckBoxSmallLettersForSubsquares.Checked, false);
					string firstSubLocator = subLocators[0];
					string lastSubLocator = subLocators[subLocators.Count - 1];

					var item = new ToolStripMenuItem("Save KML for '" + firstSubLocator + "' ... '" + lastSubLocator + "' (and do not open files)");
					{
						item.Tag = subLocators;
						item.Image = Properties.Resources.fast_forward;
					}
					items.Add(item);
				}
			}

			contextMenu.Show(control, mouseEventArgs.X, mouseEventArgs.Y);
		}

		void SaveCsvContextMenu_ItemClicked(ToolStripItemClickedEventArgs toolStripItemClickedEventArgs, Control textBoxCurrentLocator)
		{
			var item = toolStripItemClickedEventArgs.ClickedItem;
			if (item == null)
				return;

			if (!(item.Tag is List<string> locators))
				return;

			Application.DoEvents();
			SetBusyCursor();
			bool openFile = locators.Count == 1;

			foreach (string locator in locators)
				SaveCsv(locator, textBoxCurrentLocator, openFile);
		}

		void SaveKmlContextMenu_ItemClicked(ToolStripItemClickedEventArgs toolStripItemClickedEventArgs, Control textBoxCurrentLocator)
		{
			var item = toolStripItemClickedEventArgs.ClickedItem;
			if (item == null)
				return;

			if (!(item.Tag is List<string> locators))
				return;

			Application.DoEvents();
			SetBusyCursor();
			bool openFile = locators.Count == 1;

			foreach (string locator in locators)
				SaveKml(locator, textBoxCurrentLocator, openFile);
		}

		void TextBoxCsvDecimalPlaces_Validating(object sender, CancelEventArgs e)
		{
			bool success = int.TryParse(TextBoxCsvDecimalPlaces.Text, out int decimalPlaces);

			e.Cancel = !success ||
				!(decimalPlaces >= 0 && decimalPlaces <= MaxCsvDecimalPlaces);
		}

		void TextBoxCsvDecimalPlaces_Validated(object sender, EventArgs e) => RefreshComboBoxCsvAngleFormat(keepSelectedIndex: true);

		void ButtonOutputPathSelect_Click(object sender, EventArgs e)
		{
			if (folderBrowserDialogOutputPath.ShowDialog() == DialogResult.OK)
				TextBoxOutputPath.Text = folderBrowserDialogOutputPath.SelectedPath;
		}

		void CheckBoxSupergrid_CheckedChanged(object sender, EventArgs e) => SupergridCheckBoxChanged();

		void SupergridCheckBoxChanged() => CheckBoxKmlAllSuperGrids.Enabled = CheckBoxKmlSuperGrid.Checked;

		void LocatorTextBoxChanged(string locator, Control buttonLocatorEnter,
			Control buttonSaveKml, Control buttonSaveCsv, Control buttonWaypoints)
		{
			bool enableButtons = false;

			try
			{
				locator = string.IsNullOrEmpty(locator) ?
					string.Empty :
					locator.Trim();

				toolTip.SetToolTip(buttonLocatorEnter,
					"Apply locator" + NewLine + locator);

				{
					string postfix = " file" + NewLine +
						"for locator" + NewLine +
						locator + NewLine + NewLine +
						"or: right click for more options";

					toolTip.SetToolTip(buttonSaveCsv, "Save locator CSV" + postfix);
					toolTip.SetToolTip(buttonSaveKml, "Save locator KML" + postfix);
				}

				{
					string postfix = " file" + NewLine +
						"for locator" + NewLine +
						locator;
					toolTip.SetToolTip(buttonWaypoints, "Save waypoints KML" + postfix);
				}

				if (string.IsNullOrEmpty(locator))
					enableButtons = true;
				else
				{
					var dummy = new GeographicalPoint(locator, PositionInRectangle.BottomLeft);
					enableButtons = true;
				}
			}
			catch
			{
			}

			buttonSaveKml.Enabled = buttonSaveCsv.Enabled = enableButtons;
			buttonWaypoints.Enabled = !string.IsNullOrEmpty(locator) && enableButtons;
		}

		void ResetLocator(Control textBoxLocator, ListBox listBoxLocatorPrecisions)
		{
			textBoxLocator.Text = null;
			listBoxLocatorPrecisions.Items.Clear();
		}

		PositionInRectangle LocatorPositionInGrid(Control groupBox)
		{
			foreach (Control control in groupBox.Controls)
			{
				if (!(control is RadioButton radioButton))
					continue;

				if (!radioButton.Checked)
					continue;

				if (!(control.Tag is string tag))
					continue;

				var converter = TypeDescriptor.GetConverter(typeof(PositionInRectangle));
				return (PositionInRectangle)converter.ConvertFromInvariantString(tag);
			}

			throw new InvalidOperationException("No valid Position in grid.");
		}

		static int ListBoxLocatorPrecisionsIndex(string locator) => Math.Min(MaidenheadLocator.MaxPrecision - 1, locator.Length / 2 - 1);

		static string NullValue(Control control)
		{
			string text = control.Text.Trim();

			if (string.IsNullOrEmpty(text))
				return "0";

			return text;
		}

		static bool IsNegativeValue(string text)
		{
			if (string.IsNullOrEmpty(text))
				return false;

			return text.StartsWith("-");
		}

		void SaveCsv(string locator, Control textBoxCurrentLocator, bool openFile)
		{
			string fileName = "Locator_" + locator;
			string fullFileName = NormalizedPathName(TextBoxOutputPath.Text) + fileName + ".csv";

			textBoxCurrentLocator.Text = fileName;
			Refresh();
			{
				bool decimalPlacesOK = int.TryParse(TextBoxCsvDecimalPlaces.Text, out int decimalPlaces);
				if (!decimalPlacesOK)
					throw new InvalidOperationException(labelCsvDecimalPlaces.Text + ": Invalid input");

				var angleFormat = AngleFormat;

				CSV.SaveCSV(locator,
					fullFileName,
					CheckBoxCsvUtf8.Checked,
					angleFormat,
					decimalPlaces);

				if (openFile && CheckBoxOpenFile.Checked)
					StartProcess(fullFileName);
			}
			textBoxCurrentLocator.Text = null;
			Refresh();
			Application.DoEvents();
		}

		void SaveKml(string locator, Control textBoxCurrentLocator, bool openFile)
		{
			bool minLodPixelsOK = int.TryParse(TextBoxKmlMinLodPixels.Text, out int minLodPixels);
			if (!minLodPixelsOK)
				throw new InvalidOperationException(LabelKmlMinLodPixels.Text + ": Invalid input");

			string fileNamePrefix = "Locator_" + locator;
			string fullFileNamePrefix = NormalizedPathName(TextBoxOutputPath.Text) + fileNamePrefix;

			textBoxCurrentLocator.Text = fileNamePrefix;
			Refresh();
			{
				string effectiveFileName = KML.SaveMaidenheadLocatorsKML(new List<string> { locator },
					CheckBoxKmlSuperGrid.Checked, CheckBoxKmlAllSuperGrids.Checked,
					CheckBoxKmlSubGrids.Checked,
					fullFileNamePrefix,
					KmlFileFormat,
					minLodPixels,
					TextBoxKmlColorEvenPrecisions.Text, TextBoxKmlColorOddPrecisions.Text);

				if (openFile && CheckBoxOpenFile.Checked)
					StartProcess(effectiveFileName);
			}
			textBoxCurrentLocator.Text = null;
			Refresh();
			Application.DoEvents();
		}

		void SaveWaypoints(GeographicalPoint geographicalPoint, string locator,
			Control textBoxWaypointsAzimuth, Control textBoxWaypointsCount, Control textBoxWaypointsKm)
		{
			double waypointsAzimuth = double.Parse(textBoxWaypointsAzimuth.Text);
			int waypointsCount = int.Parse(textBoxWaypointsCount.Text);
			double waypointsKm = double.Parse(textBoxWaypointsKm.Text);

			var angleFormat = AngleFormat;

			int decimalPlaces = int.Parse(TextBoxCsvDecimalPlaces.Text);

			bool forceAsciiCharacters = !CheckBoxCsvUtf8.Checked;

			const string separator = " / ";

			geographicalPoint.Name = geographicalPoint.ToString(angleFormat, decimalPlaces, separator, forceAsciiCharacters);

			var waypoints = new List<GeographicalPoint>();
			{
				for (int i = 0; i <= waypointsCount; i++)
				{
					double km = waypointsKm * i;

					try
					{
						var waypoint = geographicalPoint.GetWaypointProjection(waypointsAzimuth, km);
						{
							waypoint.Name = km + " km" +
								": " +
								waypoint.ToString(angleFormat, decimalPlaces, separator, forceAsciiCharacters);
						}
						waypoints.Add(waypoint);
					}
					catch (ArgumentException)
					{
						break;
					}
				}
			}

			string fileNamePrefix = "Waypoints_" + locator;
			string fullFileNamePrefix = NormalizedPathName(TextBoxOutputPath.Text) + fileNamePrefix;

			string effectiveFileName = KML.SaveWaypointProjectionsKML(
				geographicalPoint, waypoints,
				fullFileNamePrefix,
				KmlFileFormat);

			if (CheckBoxOpenFile.Checked)
				StartProcess(effectiveFileName);
		}

		void StartProcess(string fileName) => Process.Start(fileName);

		void SetState(Control textBoxLocatorState, Exception exception)
		{
			textBoxLocatorState.Tag = exception;
			textBoxLocatorState.Text = exception.Message;

			textBoxLocatorState.BackColor = SystemColors.Highlight;
			textBoxLocatorState.ForeColor = SystemColors.HighlightText;

			textBoxLocatorState.Show();

			toolTip.SetToolTip(textBoxLocatorState, "Double click to display full error message");

			SetDefaultCursor();
		}

		void ResetState(Control textBoxLocatorState, Control linkLabelGeoHack, Control labelGeoHack, Control textBoxCurrentLocator)
		{
			textBoxLocatorState.Text = null;
			textBoxLocatorState.Tag = null;

			textBoxLocatorState.BackColor = SystemColors.Control;
			textBoxLocatorState.ForeColor = SystemColors.WindowText;

			textBoxLocatorState.Hide();

			ResetGeoHack(linkLabelGeoHack, labelGeoHack);

			textBoxCurrentLocator.Text = null;
		}

		void ShowException(Control textBoxLocatorState)
		{
			if (textBoxLocatorState.Tag is Exception exception)
				textBoxLocatorState.Text = exception.ToString();

			toolTip.SetToolTip(textBoxLocatorState, null);
		}

		void ButtonLocatorPrecisionEnter(Control buttonLocatorPrecisionEnter, Control textBoxLocator)
		{
			string text = null;
			{
				object tag = buttonLocatorPrecisionEnter.Tag;
				if (tag != null)
					text = tag as string;
			}
			textBoxLocator.Text = text;
		}

		void ListBoxLocatorPrecisionsSelectedIndexChanged(ListBox listBoxLocatorPrecisions, Control buttonLocatorPrecisionEnter)
		{
			string locator = listBoxLocatorPrecisions.SelectedItem as string;
			buttonLocatorPrecisionEnter.Tag = locator;
			toolTip.SetToolTip(buttonLocatorPrecisionEnter, "Pick locator" + NewLine + locator);
		}

		void SetGeoHackLink(Control linkLabel, Control label, GeographicalPoint geographicalPoint)
		{
			double latitude = geographicalPoint.Latitude;
			double longitude = geographicalPoint.Longitude;

			string paramsValue =
				Math.Abs(latitude).ToString().Replace(',', '.') + "_" + (latitude >= 0 ? "N" : "S") +
				"_" +
				Math.Abs(longitude).ToString().Replace(',', '.') + "_" + (longitude >= 0 ? "E" : "W");
			string url = string.Format(Settings.Default.GeoHackTemplate, paramsValue);
			SetText(linkLabel, label, url);
		}

		void ResetGeoHack(Control linkLabel, Control label) => SetText(linkLabel, label, null);

		void SetText(Control linkLabel, Control label, string text)
		{
			linkLabel.Text = text;
			toolTip.SetToolTip(linkLabel, text);
			label.Visible = !string.IsNullOrEmpty(text);
		}

		static string Formatted(double value, int decimalPlaces) => string.Format("{0:0." + new string('0', decimalPlaces) + "}", value);

		void GeographicalPointUpdated(GeographicalPoint geographicalPoint,
			Control textBoxDND, Control textBoxDED,
			Control textBoxDMND, Control textBoxDMNM, Control textBoxDMED, Control textBoxDMEM,
			Control textBoxDMSND, Control textBoxDMSNM, Control textBoxDMSNS, Control textBoxDMSED, Control textBoxDMSEM, Control textBoxDMSES)
		{
			//D
			{
				int decimalPlaces = MaxCsvDecimalPlaces + 2;

				{
					Utilities.DegreesToD(geographicalPoint.Latitude, out bool isNegative, out _);
					textBoxDND.Text = GeographicalPoint.GetSignSymbol(isNegative) +
						Formatted(Math.Abs(geographicalPoint.Latitude), decimalPlaces);
				}
				{
					Utilities.DegreesToD(geographicalPoint.Longitude, out bool isNegative, out _);
					textBoxDED.Text = GeographicalPoint.GetSignSymbol(isNegative) +
						Formatted(Math.Abs(geographicalPoint.Longitude), decimalPlaces);
				}
			}

			//DM
			{
				int decimalPlaces = MaxCsvDecimalPlaces + 2;

				{
					Utilities.DegreesToDM(geographicalPoint.Latitude, out bool isNegative, out int degrees, out double minutes);
					Utilities.Round(decimalPlaces, ref degrees, ref minutes);
					textBoxDMND.Text =
						GeographicalPoint.GetSignSymbol(isNegative) +
						degrees;
					textBoxDMNM.Text = Formatted(minutes, decimalPlaces);
				}
				{
					Utilities.DegreesToDM(geographicalPoint.Longitude, out bool isNegative, out int degrees, out double minutes);
					Utilities.Round(decimalPlaces, ref degrees, ref minutes);
					textBoxDMED.Text =
						GeographicalPoint.GetSignSymbol(isNegative) +
						degrees;
					textBoxDMEM.Text = Formatted(minutes, decimalPlaces);
				}
			}

			//DMS
			{
				int decimalPlaces = MaxCsvDecimalPlaces;

				{
					Utilities.DegreesToDMS(geographicalPoint.Latitude, out bool isNegative, out int degrees, out int minutes, out double seconds);
					Utilities.Round(decimalPlaces, ref degrees, ref minutes, ref seconds);
					textBoxDMSND.Text =
						GeographicalPoint.GetSignSymbol(isNegative) +
						degrees;
					textBoxDMSNM.Text = minutes.ToString();
					textBoxDMSNS.Text = Formatted(seconds, decimalPlaces);
				}
				{
					Utilities.DegreesToDMS(geographicalPoint.Longitude, out bool isNegative, out int degrees, out int minutes, out double seconds);
					Utilities.Round(decimalPlaces, ref degrees, ref minutes, ref seconds);
					textBoxDMSED.Text =
						GeographicalPoint.GetSignSymbol(isNegative) +
						degrees;
					textBoxDMSEM.Text = minutes.ToString();
					textBoxDMSES.Text = Formatted(seconds, decimalPlaces);
				}
			}

			GeographicalPointsUpdated();
		}

		void GeographicalPointsUpdated()
		{
			if (GeographicalPoint1 == null || GeographicalPoint2 == null)
				return;

			const int decimalPlaces = 6;

			{
				double distanceKM = GeographicalPoint1.GetDistance(GeographicalPoint2);
				double distanceStatuteMiles = distanceKM / GeographicalPoint.StatuteMilesToKilometres;
				double distanceNauticalMiles = distanceKM / GeographicalPoint.NauticalMilesToKilometres;

				textBoxDistanceKM.Text = Formatted(distanceKM, decimalPlaces);
				textBoxDistanceStatuteMiles.Text = Formatted(distanceStatuteMiles, decimalPlaces);
				textBoxDistanceNauticalMiles.Text = Formatted(distanceNauticalMiles, decimalPlaces);
			}

			{
				double azimuth12 = GeographicalPoint1.GetAzimuth(GeographicalPoint2);
				double azimuth21 = GeographicalPoint2.GetAzimuth(GeographicalPoint1);

				if (Math.Round(azimuth12, decimalPlaces) == 360)
					azimuth12 = 0;

				if (Math.Round(azimuth21, decimalPlaces) == 360)
					azimuth21 = 0;

				textBoxAzimuth12.Text = Formatted(azimuth12, decimalPlaces);
				textBoxAzimuth21.Text = Formatted(azimuth21, decimalPlaces);
			}
		}

		static string NormalizedPathName(string oldPath)
		{
			oldPath = oldPath.Trim();

			if (string.IsNullOrEmpty(oldPath))
				return oldPath;

			char directorySeparatorChar = System.IO.Path.DirectorySeparatorChar;
			int oldLength = oldPath.Length;
			int lastIndex = oldPath.LastIndexOf(directorySeparatorChar);

			if (lastIndex == oldLength - 1)
				return oldPath;

			return oldPath + directorySeparatorChar;
		}
	}
}